import { injectable, inject } from 'inversify';
import { IConnection } from 'vscode-languageserver';
import {
    CompletionList, Hover,
    SymbolInformation, DocumentSymbolParams, TextEdit, DocumentHighlight, DocumentLink,
} from 'vscode-languageserver-types';
import { getLanguageService, LanguageService, HTMLDocument } from 'vscode-html-languageservice';
import {
    TextDocumentPositionParams, DocumentRangeFormattingParams,
} from 'vscode-base-languageclient';
import { TextDocuments } from 'vscode-languageserver';
import { uriToPath, TextDocument, convertChangesRange } from '@workbench-stack/language-core';
import { logger } from '@cdm-logger/server';

@injectable()
export class HtmlService {
    private openedDocumentUris: Map<string, any> = new Map<string, any>();

    protected readonly htmlService: LanguageService = getLanguageService();
    public readonly documents;

    public onInitialize = () => {
        logger.debug('The html server is initialized.');
        return {
            capabilities: {
                completionProvider: true,
                hoverProvider: true,
                documentHighlightProvider: true,
                documentRangeFormattingProvider: false,
                documentLinkProvider: true,
                documentSymbolProvider: true,
                signatureHelpProvider: { triggerCharacters: ['('] },
                referencesProvider: true,
                colorProvider: true,
            },
        };
    }

    public onDidOpenTextDocument = params => {
        logger.debug('onDidOpenTextDocument  html');
        const path = uriToPath(params.textDocument.uri);
        const document = new TextDocument(params.textDocument);
        this.openedDocumentUris.set(params.textDocument.uri, document);
    }

    public onDidChangeTextDocument = params => {
        logger.debug('onDidChangeTextDocument html');
        const document = this.getOpenDocument(params.textDocument.uri);
        const changes = convertChangesRange(params.contentChanges);
        document.apply(changes, params.textDocument.version);
    }

    public onDidCloseTextDocument = async (params) => {
        const path = uriToPath(params.textDocument.uri);
        this.openedDocumentUris.delete(params.textDocument.uri);
    }

    public onDocumentFormatting(params: DocumentRangeFormattingParams): TextEdit[] {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        return this.htmlService.format(document, params.range, params.options);
    }

    public onDocumentSymbol(params: DocumentSymbolParams): SymbolInformation[] {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const htmlDocument = this.getHTMLDocument(document);
        return this.htmlService.findDocumentSymbols(document, htmlDocument);
    }

    public onDocumentHighlight(params): DocumentHighlight[] {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const htmlDocument = this.getHTMLDocument(document);
        return this.htmlService.findDocumentHighlights(document, params.position, htmlDocument);
    }

    public onDocumentLinks(params): DocumentLink[] {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const htmlDocument = this.getHTMLDocument(document);
        const res = this.htmlService.findDocumentLinks(document, htmlDocument['_content']);
        return res.map(item => {
            return {
                ... item,
                url: item.target,
            };
        });
    }

    public onHover(params: TextDocumentPositionParams): Hover | null {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const htmlDocument = this.getHTMLDocument(document);
        return this.htmlService.doHover(document, params.position, htmlDocument);
    }

    public onCompletion(params: TextDocumentPositionParams): CompletionList {
        const document = this.getOpenDocument(params.textDocument.uri) as any;
        const htmlDocument = this.getHTMLDocument(document);
        const res = this.htmlService.doComplete(document, params.position, htmlDocument);
        res.items = res.items.map(item => {
            if (!item.detail) {
                item.detail = '';
            }
            return item;
        });
        return res;
    }

    public getHTMLDocument(document: any): HTMLDocument {
        return this.htmlService.parseHTMLDocument(document);
    }

    private getOpenDocument(uri: string): TextDocument {
        const doc = this.openedDocumentUris.get(uri);
        if (doc) {
            return doc;
        } else {
            throw new Error('Document ' + uri + ' has not been opened.');
        }
    }

}
